@@ -17,6 +17,7 @@ use LaTeXML::Global;
 use LaTeXML::Common::Object;
 use LaTeXML::Common::Error;
 use LaTeXML::Core::Token;
+use LaTeXML::Core::Tokens;
 use LaTeXML::Core::Gullet;
 use LaTeXML::Core::Box;
 use LaTeXML::Core::Comment;
@@ -57,8 +58,8 @@ sub getGullet {
   return $$self{gullet}; }
 
 sub getLocator {
-  my ($self, @args) = @_;
-  return $$self{gullet}->getLocator(@args); }
+  my ($self) = @_;
+  return $$self{gullet}->getLocator; }
 
 sub getBoxingLevel {
   my ($self) = @_;
@@ -72,6 +73,7 @@ sub getBoxingLevel {
 
 sub digestNextBody {
   my ($self, $terminal) = @_;
+  no warnings 'recursion';
   my $startloc  = $self->getLocator;
   my $initdepth = scalar(@{ $$self{boxing} });
   my $token;
@@ -83,7 +85,7 @@ sub digestNextBody {
   Warn('expected', $terminal, $self, "body should have ended with '" . ToString($terminal) . "'",
     "current body started at " . ToString($startloc))
     if $terminal && !Equals($token, $terminal);
-  push(@LaTeXML::LIST, LaTeXML::Core::List->new()) unless $token;  # Dummy `trailer' if none explicit.
+  push(@LaTeXML::LIST, Box()) unless $token;                      # Dummy `trailer' if none explicit.
   return @LaTeXML::LIST; }
 
 # Digest a list of tokens independent from any current Gullet.
@@ -91,6 +93,7 @@ sub digestNextBody {
 # Returns a List containing the digested material.
 sub digest {
   my ($self, $tokens) = @_;
+  no warnings 'recursion';
   return unless defined $tokens;
   return
     $$self{gullet}->readingFromMouth(LaTeXML::Core::Mouth->new(), sub {
@@ -121,18 +124,55 @@ sub digest {
 # Returns a list of boxes/whatsits.
 my $MAXSTACK = 200;    # [CONSTANT]
 
+# Overly complex, but want to avoid recursion/stack
+my @absorbable_cc = (    # [CONSTANT]
+  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 1, 0);
+
 sub invokeToken {
   my ($self, $token) = @_;
+  no warnings 'recursion';
+INVOKE:
   push(@{ $$self{token_stack} }, $token);
   if (scalar(@{ $$self{token_stack} }) > $MAXSTACK) {
     Fatal('internal', '<recursion>', $self,
       "Excessive recursion(?): ",
       "Tokens on stack: " . join(', ', map { ToString($_) } @{ $$self{token_stack} })); }
-  my @result = $self->invokeToken_internal($token);
+  local $LaTeXML::CURRENT_TOKEN = $token;
+  my @result  = ();
+  my $meaning = $STATE->lookupDigestableDefinition($token);
+
+  if ($meaning->isaToken) {    # Common case
+    my $cc = $meaning->getCatcode;
+    if ($cc == CC_CS) {
+      @result = $self->invokeToken_undefined($token); }
+    elsif ($absorbable_cc[$cc]) {
+      @result = $self->invokeToken_simple($token, $meaning); }
+    else {
+      Error('misdefined', $token, $self,
+        "The token " . Stringify($token) . " should never reach Stomach!");
+      @result = $self->invokeToken_simple($token, $meaning); } }
+  # A math-active character will (typically) be a macro,
+  # but it isn't expanded in the gullet, but later when digesting, in math mode (? I think)
+  elsif ($meaning->isExpandable) {
+    my $gullet = $$self{gullet};
+    $gullet->unread(@{ $meaning->invoke($gullet) || [] });
+    $token = $gullet->readXToken();    # replace the token by it's expansion!!!
+    pop(@{ $$self{token_stack} });
+    goto INVOKE; }
+  elsif ($meaning->isaDefinition) {    # Otherwise, a normal primitive or constructor
+    @result = $meaning->invoke($self);
+    $STATE->clearPrefixes unless $meaning->isPrefix; }    # Clear prefixes unless we just set one.
+  else {
+    Fatal('misdefined', $meaning, $self,
+      "The object " . Stringify($meaning) . " should never reach Stomach!"); }
   if ((scalar(@result) == 1) && (!defined $result[0])) {
-    @result = (); }    # Just paper over the obvious thing.
-  if (my ($x) = grep { !$_->isaBox } @result) {
-    Fatal('misdefined', $x, $self, "Expected a Box|List|Whatsit, but got '" . Stringify($x) . "'"); }
+    @result = (); }                                       # Just paper over the obvious thing.
+  Fatal('misdefined', $token, $self,
+    "Execution yielded non boxes",
+    "Returned " . join(',', map { "'" . Stringify($_) . "'" }
+        grep { (!ref $_) || (!$_->isaBox) } @result))
+    if grep { (!ref $_) || (!$_->isaBox) } @result;
+
   pop(@{ $$self{token_stack} });
   return @result; }
 
@@ -147,40 +187,30 @@ sub makeError {
   $document->setNode($savenode) if $savenode;
   return; }
 
-my @forbidden_cc = (    # [CONSTANT]
-  1, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 1);
-
-sub invokeToken_internal {
-  my ($self, $token) = @_;
-  local $LaTeXML::CURRENT_TOKEN = $token;
-  my $meaning = ($token->isExecutable
-      || ($STATE->lookupValue('IN_MATH')
-      && (($STATE->lookupMathcode($token->getString) || 0) == 0x8000))
-    ? $STATE->lookupMeaning_internal($token)
-    : $token);
-
-  if (!defined $meaning) {    # Supposedly executable token, but no definition!
-    return $self->invokeToken_undefined($token); }
-  elsif ($meaning->isaToken) {    # Common case
-    return $self->invokeToken_simple($token, $meaning); }
-  elsif ($meaning->isaDefinition) {
-    return $self->invokeToken_definition($token, $meaning); }
-  else {
-    Fatal('misdefined', $meaning, $self,
-      "The object " . Stringify($meaning) . " should never reach Stomach!");
-    return; } }
-
 sub invokeToken_undefined {
   my ($self, $token) = @_;
   my $cs = $token->getCSName;
   $STATE->noteStatus(undefined => $cs);
-  Error('undefined', $token, $self, "The token " . Stringify($token) . " is not defined.");
   # To minimize chatter, go ahead and define it...
-  $STATE->installDefinition(LaTeXML::Core::Definition::Constructor->new($token, undef,
-      sub { makeError($_[0], 'undefined', $cs); }),
-    'global');
-  # and then invoke it.
-  return $self->invokeToken($token); }
+  if ($cs =~ /^\\if(.*)$/) {    # Apparently an \ifsomething ???
+    my $name = $1;
+    Error('undefined', $token, $self, "The token " . Stringify($token) . " is not defined.",
+      "Defining it now as with \\newif");
+    $STATE->installDefinition(LaTeXML::Core::Definition::Expandable->new(
+        T_CS('\\' . $name . 'true'), undef, '\let' . $cs . '\iftrue'));
+    $STATE->installDefinition(LaTeXML::Core::Definition::Expandable->new(
+        T_CS('\\' . $name . 'false'), undef, '\let' . $cs . '\iffalse'));
+    LaTeXML::Package::Let($token, T_CS('\iffalse'));
+    $self->getGullet->unread($token);    # Retry
+    return; }
+  else {
+    Error('undefined', $token, $self, "The token " . Stringify($token) . " is not defined.",
+      "Defining it now as <ltx:ERROR/>");
+    $STATE->installDefinition(LaTeXML::Core::Definition::Constructor->new($token, undef,
+        sub { makeError($_[0], 'undefined', $cs); }),
+      'global');
+    # and then invoke it.
+    return $self->invokeToken($token); } }
 
 sub invokeToken_simple {
   my ($self, $token, $meaning) = @_;
@@ -188,49 +218,20 @@ sub invokeToken_simple {
   my $font = $STATE->lookupValue('font');
   $STATE->clearPrefixes;    # prefixes shouldn't apply here.
   if ($cc == CC_SPACE) {
-    if (($STATE->lookupValue('IN_MATH') || $STATE->lookupValue('inPreamble'))) {
+    if ($STATE->lookupValue('IN_MATH')) {    # (but in Preamble, OK ?)
       return (); }
     else {
       return Box($meaning->getString, $font, $$self{gullet}->getLocator, $meaning); } }
-  elsif ($cc == CC_COMMENT) {    # Note: Comments need char decoding as well!
+  elsif ($cc == CC_COMMENT) {                # Note: Comments need char decoding as well!
     my $comment = LaTeXML::Package::FontDecodeString($meaning->getString, undef, 1);
     # However, spaces normally would have be digested away as positioning...
     my $badspace = pack('U', 0xA0) . "\x{0335}";    # This is at space's pos in OT1
     $comment =~ s/\Q$badspace\E/ /g;
     return LaTeXML::Core::Comment->new($comment); }
-  elsif ($forbidden_cc[$cc]) {
-    Fatal('misdefined', $token, $self,
-      "The token " . Stringify($token) . " should never reach Stomach!");
-    return; }
   else {
     return Box(LaTeXML::Package::FontDecodeString($meaning->getString, undef, 1),
       undef, undef, $meaning); } }
 
-sub invokeToken_definition {
-  my ($self, $token, $meaning) = @_;
-  # A math-active character will (typically) be a macro,
-  # but it isn't expanded in the gullet, but later when digesting, in math mode (? I think)
-  if ($meaning->isExpandable) {
-    my $gullet = $$self{gullet};
-    $gullet->unread($meaning->invoke($$self{gullet}));
-    if (my $replacement = $gullet->readXToken()) {
-      return $self->invokeToken($replacement); }    # recurse!!
-    return; }                                       # fell off?
-  else {                                            # Otherwise, a normal primitive or constructor
-    if ($token->equals(T_CS('\par') && $STATE->lookupValue('inPreamble'))) {
-      return (); }
-    my @boxes = $meaning->invoke($self);
-    if ((scalar(@boxes) == 1) && (!defined $boxes[0])) {
-      @boxes = (); }                                # Just paper over the obvious thing.
-    my @err = grep { (!ref $_) || (!$_->isaBox) } @boxes;
-    Fatal('misdefined', $token, $self,
-      "Execution yielded non boxes",
-      "Returned " . join(',', map { "'" . Stringify($_) . "'" }
-          grep { (!ref $_) || (!$_->isaBox) } @boxes))
-      if grep { (!ref $_) || (!$_->isaBox) } @boxes;
-    $STATE->clearPrefixes unless $meaning->isPrefix;    # Clear prefixes unless we just set one.
-    return @boxes; } }
-
 # Regurgitate: steal the previously digested boxes from the current level.
 sub regurgitate {
   my ($self) = @_;
@@ -286,6 +287,12 @@ sub currentFrameMessage {
 #======================================================================
 # Grouping pushes a new stack frame for binding definitions, etc.
 #======================================================================
+# Originally, we only treated math vs text "modes", which are correlated
+# to grouping (somehow). But we'll gradually need to incorporate all
+# the horizontal/vertical modes, which are NOT correlated to grouping,
+# although they do operate on a stack.
+# So, we should NOT generate errors when the grouping clashes with modes
+# (until we can get it properly sorted).
 
 # if $nobox is true, inhibit incrementing the boxingLevel
 sub bgroup {
@@ -295,11 +302,11 @@ sub bgroup {
 
 sub egroup {
   my ($self) = @_;
-  if ($STATE->isValueBound('MODE', 0)    # Last stack frame was a mode switch!?!?!
-    || $STATE->lookupValue('groupNonBoxing')) {    # or group was opened with \begingroup
+  if (    ##$STATE->isValueBound('MODE', 0) ||    # Last stack frame was a mode switch!?!?!
+    $STATE->lookupValue('groupNonBoxing')) {    # or group was opened with \begingroup
     Error('unexpected', $LaTeXML::CURRENT_TOKEN, $self, "Attempt to close boxing group",
       $self->currentFrameMessage); }
-  else {    # Don't pop if there's an error; maybe we'll recover?
+  else {                                        # Don't pop if there's an error; maybe we'll recover?
     popStackFrame($self, 0); }
   return; }
 
@@ -310,11 +317,11 @@ sub begingroup {
 
 sub endgroup {
   my ($self) = @_;
-  if ($STATE->isValueBound('MODE', 0)    # Last stack frame was a mode switch!?!?!
-    || !$STATE->lookupValue('groupNonBoxing')) {    # or group was opened with \bgroup
+  if (    ##$STATE->isValueBound('MODE', 0) ||    # Last stack frame was a mode switch!?!?!
+    !$STATE->lookupValue('groupNonBoxing')) {    # or group was opened with \bgroup
     Error('unexpected', $LaTeXML::CURRENT_TOKEN, $self, "Attempt to close non-boxing group",
       $self->currentFrameMessage); }
-  else {    # Don't pop if there's an error; maybe we'll recover?
+  else {                                         # Don't pop if there's an error; maybe we'll recover?
     popStackFrame($self, 1); }
   return; }
 
@@ -322,9 +329,11 @@ sub endgroup {
 # Mode (minimal so far; math vs text)
 # Could (should?) be taken up by Stomach by building horizontal, vertical or math lists ?
 
-sub beginMode {
+# This sets the mode without doing any grouping (NOR does it stack the modes!!)
+# Useful for environments, where the group has already been established.
+# (presumably, in the long run, modes & groups should be much less coupled)
+sub setMode {
   my ($self, $mode) = @_;
-  $self->pushStackFrame;    # Effectively bgroup
   my $prevmode = $STATE->lookupValue('MODE');
   my $ismath   = $mode =~ /math$/;
   $STATE->assignValue(MODE    => $mode,   'local');
@@ -335,11 +344,22 @@ sub beginMode {
     # When entering math mode, we set the font to the default math font,
     # and save the text font for any embedded text.
     $STATE->assignValue(savedfont => $curfont, 'local');
-    $STATE->assignValue(font => $STATE->lookupValue('mathfont')->merge(color => $curfont->getColor), 'local');
-    $STATE->assignValue(mathstyle => ($mode =~ /^display/ ? 'display' : 'text'), 'local'); }
+    $STATE->assignValue(font => $STATE->lookupValue('mathfont')->merge(
+        color => $curfont->getColor, background => $curfont->getBackground,
+        size => $curfont->getSize,
+        mathstyle => ($mode =~ /^display/ ? 'display' : 'text')), 'local'); }
   else {
-# When entering text mode, we should set the font to the text font in use before the math (but inherit color!).
-    $STATE->assignValue(font => $STATE->lookupValue('savedfont')->merge(color => $curfont->getColor), 'local'); }
+    # When entering text mode, we should set the font to the text font in use before the math
+    # but inherit color and size
+    $STATE->assignValue(font => $STATE->lookupValue('savedfont')->merge(
+        color => $curfont->getColor, background => $curfont->getBackground,
+        size => $curfont->getSize), 'local'); }
+  return; }
+
+sub beginMode {
+  my ($self, $mode) = @_;
+  $self->pushStackFrame;    # Effectively bgroup
+  $self->setMode($mode);
   return; }
 
 sub endMode {
